from thercy.constants import PartType
from thercy.state import StatePoint

from .base_part import BasePart, Connection


class Turbine(BasePart):
    def __init__(self, label, p_out, eta=1.0, connections=None):
        """
        Parameters
        ----------
        label : str
        p_out : float
        eta : float
        connections : list[Connection]

        """
        super().__init__(
            label,
            PartType.TURBINE,
            connections,
        )

        self._p_out = p_out
        self._eta = eta
        self._deltaH = 0.0

    @property
    def deltaH(self):
        return self._deltaH

    def solve(self, inlets: dict[str, StatePoint]):
        outlets = {}

        inlet_label, inlet_state = next(iter(inlets.items()))
        outlet_state = inlet_state.clone()

        outlet_state['P'] = self._p_out
        outlet_state['S'] = inlet_state['S']
        outlet_state.properties('P', 'S')

        outlet_state['H'] = inlet_state['H'] - self._eta * (inlet_state['H'] - outlet_state['H'])
        outlet_state.properties('P', 'H')

        self._deltaH = outlet_state['H'] - inlet_state['H']

        for outlet in self.get_outlets(inlet_label):
            outlets[outlet.label] = outlet_state.clone()

        return outlets



from thercy.constants import PartType
from thercy.state import StatePoint

from .base_part import BasePart, Connection


class HeaterClosedReal(BasePart):
    def __init__(self, label, connections=None):
        """
        Parameters
        ----------
        label : str
        connections : list[Connection]

        """
        super().__init__(
            label,
            PartType.REHEATER_OPEN,
            connections,
        )

        self._deltaH = 0.0

    @property
    def deltaH(self):
        return self._deltaH

    def solve(self, inlets: dict[str, StatePoint]):
        inlet_hp: dict[str, StatePoint] = {}
        inlets_lp: dict[str, StatePoint] = {}
        pressures = []

        for label, state in inlets.items():
            pressures.append(state['P'])

        pressure_max = max(pressures)
        for label, state in inlets.items():
            if state['P'] == pressure_max:
                inlet_hp[label] = state
            else:
                inlets_lp[label] = state

        outlet_hp_state = StatePoint(next(iter(inlets.values())).fluid)
        outlet_lp_state = StatePoint(next(iter(inlets.values())).fluid)
        outlets = {}

        partial_y_lp = 0.0
        partial_p_lp = 0.0
        partial_h_lp = 0.0
        for inlet in inlets_lp.values():
            partial_y_lp += inlet['Y']
            partial_p_lp += inlet['Y'] * inlet['P']
            partial_h_lp += inlet['Y'] * inlet['H']

        outlet_lp_state['P'] = partial_p_lp / partial_y_lp
        outlet_lp_state['Q'] = 0.0
        outlet_lp_state.properties('P', 'Q')
        # outlet_lp_state['Y'] = partial_y_lp

        dH = partial_y_lp * outlet_lp_state['H'] - partial_h_lp

        partial_y_hp = 0.0
        partial_p_hp = 0.0
        partial_h_hp = 0.0
        for inlet in inlet_hp.values():
            partial_y_hp += inlet['Y']
            partial_p_hp += inlet['Y'] * inlet['P']
            partial_h_hp += inlet['Y'] * inlet['H']

        outlet_hp_state['P'] = partial_p_hp / partial_y_hp
        outlet_hp_state['H'] = partial_h_hp + dH / partial_y_hp
        outlet_hp_state.properties('P', 'H')
        # outlet_hp_state['Y'] = partial_y_hp

        for outlet in self.get_outlets(next(iter(inlet_hp.keys()))):
            outlets[outlet.label] = outlet_hp_state.clone()

        for outlet in self.get_outlets(next(iter(inlets_lp.keys()))):
            outlets[outlet.label] = outlet_lp_state.clone()

        return outlets



from thercy.constants import PartType
from thercy.state import StatePoint

from .base_part import BasePart, Connection


class HeaterOpen(BasePart):
    def __init__(self, label, connections=None):
        """
        Parameters
        ----------
        label : str
        connections : list[Connection]

        """
        super().__init__(
            label,
            PartType.REHEATER_OPEN,
            connections,
        )

        self._deltaH = 0.0

    @property
    def deltaH(self):
        return self._deltaH

    def solve(self, inlets: dict[str, StatePoint]):
        outlets = {}
        outlet_state = StatePoint(next(iter(inlets.values())).fluid)

        partial_p = 0.0
        partial_y = 0.0
        for inlet in inlets.values():
            partial_p += inlet['Y'] * inlet['P']
            partial_y += inlet['Y']

        outlet_state['P'] = partial_p / partial_y
        outlet_state['Q'] = 0.0
        outlet_state.properties('P', 'Q')

        # outlet_state['Y'] = partial_y

        # Only one outlet
        for outlet in self.get_outlets(next(iter(inlets.keys()))):
            outlets[outlet.label] = outlet_state.clone()

        return outlets



__thercy_phys_submodules__ = {'cycles', 'state', }

__all__ = [__thercy_phys_submodules__]


def __getattr__(attr):
    if attr == 'cycles':
        import thercy.cycles as cycles
        return cycles
    elif attr == 'state':
        import thercy.state as state
        return state



import numpy as np
from CoolProp.Plots import StateContainer

from thercy.constants import Property, PropertyInfo
from thercy.utils import list_like

from .point import StatePoint


class StateCycle:
    """
    Thermodynamic state cycle composed of multiple state points.

    Properties
    ----------
    fluid : str
        Fluid composition or mixture name.
    data : dict
        Dictionary containing keyed state points of the cycle.

    """
    def __init__(self, fluid, data=None):
        """
        Initialize a StateCycle instance.

        Parameters
        ----------
        fluid : str
            Fluid composition or mixture name.
        data : dict[any: StatePoint], optional
            Dictionary containing state points of the cycle. Default: {}

        """
        if data is None:
            data = {}

        self._data: dict[any: StatePoint] = data
        self._fluid: str = fluid

    def __len__(self):
        """Get the number of state points in the cycle."""
        return len(self._data)

    def __iter__(self):
        """Iterator over state points in the cycle."""
        return iter(self._data)

    def __str__(self):
        """String representation of the cycle's state points."""
        out = ''

        # Retrieve the keys of the first StatePoint to get property labels
        row = [f"{'State':>5s}"]
        for prop in Property:
            label = f"{PropertyInfo.label(prop)} ({PropertyInfo.unit(prop)})"
            row.append(f"{label:>16s}")
        out += '  '.join(row) + '\n'

        for k in self._data.keys():
            row = [f"{str(k):>5s}"]
            for prop in Property:
                value = self._data[k][prop.name]
                if value is not None:
                    row.append(f"{value:16.3f}")
                else:
                    row.append(f"{'-':>16s}")
            out += '  '.join(row) + '\n'

        return out

    def __getitem__(self, key):
        """
        Get a state point by keys.

        Parameters
        ----------
        key : any | tuple
            Key to access the state point.

        Returns
        -------
        StatePoint | float
            State point corresponding to the key.

        Raises
        ------
        IndexError if the index has an invalid length.

        """
        if list_like(key):
            len_var = len(key)
            if len_var == 1:
                return self._get_point(key[0])
            elif len_var == 2:
                return self._get_float(key)
            else:
                raise IndexError("Index with invalid length.")

        return self._get_point(key)

    def _get_float(self, key):
        """
        Get a state point property value by key pair.

        Parameters
        ----------
        key : tuple
            Key pair to access the state point.

        Returns
        ------
        value : float

        """
        return self._data[key[0]][key[1]]

    def _get_point(self, key):
        """
        Get a state point value by key.

        Parameters
        ----------
        key : any
            Key to access the state point.

        Returns
        ------
        value : StatePoint

        """
        return self._data[key]

    def __setitem__(self, key, value):
        """
        Set a state point by key or key pair.

        Parameters
        ----------
        key : any | tuple
            Key to access the state point.
        value : StatePoint | float
            State point or float value to assign.

        Raises
        ------
        IndexError if the index has an invalid length.

        """
        if list_like(key):
            len_var = len(key)
            if len_var == 1:
                self._set_point(key[0], value)
            elif len_var == 2:
                self._set_float(key, value)
            else:
                raise IndexError("Index with invalid length.")
        else:
            self._set_point(key, value)

    def _set_float(self, key, value):
        """
        Set a state point value by key pair.

        Parameters
        ----------
        key : tuple
            Key pair to access the state point.
        value : float
            Float value to assign.

        Raises
        ------
        TypeError if the value type is not numeric.

        """
        if not isinstance(value, (int, float)):
            raise TypeError('Value is not a numeric type')

        if key[0] not in self._data:
            self._data[key[0]] = StatePoint(self._fluid)

        self._data[key[0]][key[1]] = value

    def _set_point(self, key, value):
        """
        Set a state point by key.

        Parameters
        ----------
        key : any
            Key to access the state point.
        value : StatePoint
            State point to assign.

        Raises
        ------
        TypeError if the value type is invalid.

        """
        if not isinstance(value, StatePoint):
            raise TypeError('Value is not a StatePoint')

        self._data[key] = value

    @property
    def fluid(self):
        """Get the fluid composition or mixture name."""
        return self._fluid

    @property
    def first(self):
        """Get the first state point in the cycle."""
        if self._data:
            return next(iter(self._data.values()))

    @property
    def last(self):
        """Get the last state point in the cycle."""
        if self._data:
            return next(iter(reversed(self._data.values())))

    def to_state_container(self):
        """
        Convert the cycle's data to CoolProp.Plots.StateContainer.

        Returns
        -------
        CoolProp.Plots.StateContainer
            State container.

        """
        container = StateContainer()

        for i, k in enumerate(self._data.keys()):
            for prop in Property:
                if prop.name not in ['Y']:
                    container[i, prop.name] = self[k, prop.name]

        return container

    def calculate_properties(self):
        """Calculate properties for all state points in the cycle."""
        for sp in self._data.values():
            sp.properties()

    def constant_properties(self, key1, key2, tol=1e-7):
        """
        Get properties with constant values between two state points.

        Parameters
        ----------
        key1 : any
            Key of the first state point.
        key2 : any
            Key of the second state point.
        tol : float, optional
            Tolerance for constant property comparison. Default: 1e-4

        Returns
        -------
        list[Property]
            List of constant properties.

        """
        return [prop for prop in Property if abs(self[key1, prop.name] - self[key2, prop.name]) <= tol]

    def cloud_points(self, n=50, close_envelope=False, precise=False):
        """
        Generate cloud points between consecutive state points.

        Parameters
        ----------
        n : int, optional
            Number of points between each pair of state points.
            Default: 50
        close_envelope : bool, optional
            Close the cycle envelope by connecting the last state to the
            first state. Default: False
        precise : bool, optional
            Use a more precise version of interpolation by enthalpy
            instead of temperature, slower. Default: False

        Returns
        -------
        StateCycle
            State cycle containing the cloud points.

        """
        cloud = StateCycle(self._fluid)

        for i, (key1, state1) in enumerate(self._data.items()):
            # Close the cycle envelope
            if i + 1 < len(self._data):
                key2, state2 = list(self._data.items())[i + 1]
            elif close_envelope:
                key2, state2 = list(self._data.items())[0]
            else:
                break

            prop_x = next((prop for prop in self.constant_properties(key1, key2)
                           if prop.name not in ('S', 'Q', 'Y')), 'H' if precise else 'T')

            s_diff = state2['S'] - state1['S']
            y_diff = state2[prop_x] - state1[prop_x]

            s = np.linspace(state1['S'], state2['S'], n - 1)
            s = np.append(s, state2['S'])

            x = (y_diff / s_diff) * (s - state1['S']) + state1[prop_x] if abs(s_diff) > 1e-7 \
                else np.linspace(state1[prop_x], state2[prop_x], n)

            for j in range(n):
                cloud_key = j + i * n
                cloud[cloud_key, 'S'] = s[j]
                cloud[cloud_key, prop_x] = x[j]
                cloud[cloud_key].properties()

        return cloud



import numpy as np

from thercy.utils import list_like
from .cycle import StateCycle
from .point import StatePoint


class StateGraph:
    """
    Graph representing a thermodynamic state system composed of
    interconnected parts.

    Properties
    ----------
    fluid : str
        Fluid composition or mixture name.
    states : StateCycle
        State cycle containing state points corresponding to edges.

    """
    def __init__(self, fluid):
        """
        Initialize a StateGraph instance.

        Parameters
        ----------
        fluid : str
            Fluid composition or mixture name.

        """
        self._nodes: dict[str, any] = {}
        self._edges: list[(str, str)] = []
        self._edges_map: dict[(str, str): int] = {}

        self.states: StateCycle = StateCycle(fluid)

    def __iter__(self):
        """Iterator over graph nodes."""
        return iter(self._nodes.values())

    def __len__(self):
        """Get the number of nodes in the graph."""
        return len(self._nodes)

    def __str__(self):
        """String representation of the state cycle."""
        return str(self.states)

    def __getitem__(self, key):
        """
        Get a node by key.

        Parameters
        ----------
        key : str
            Node key.

        Returns
        -------
        any
            Node corresponding to the key.

        """
        return self._nodes[key]

    @property
    def points(self):
        """Get the number of edges in the graph."""
        return len(self._edges)

    @property
    def nodes(self):
        """Get the nodes in the graph."""
        return self._nodes

    @property
    def edges(self):
        """Get the edges in the graph."""
        return self._edges

    def get_edge_index(self, edge):
        """
        Get the index of an edge.

        Parameters
        ----------
        edge : tuple
            Edge to find.

        Returns
        -------
        int
            Index of the edge.

        """
        return self._edges_map[edge]

    def get_state(self, edge):
        return self.states[self.get_edge_index(edge)]

    def get_outlets(self, key, inlet_key=None, index=False):
        """
        Get outlet keys connected to a node.

        Parameters
        ----------
        key : str
            Key of the node.
        inlet_key : str, optional
            Inlet key to filter outlets. Default: None
        index : bool, optional
            If True, return indices instead of keys. Default: False

        Returns
        -------
        list[str | int]
            List of outlet keys connected to the node.

        """
        if inlet_key is None:
            return {i if index else edge[1] for i, edge in enumerate(self._edges) if edge[0] == key}

        if index:
            raise NotImplementedError()

        outlets = []
        for conn in self._nodes[key].connections:
            for inlet in conn.inlets:
                if inlet.label == inlet_key:
                    for outlet in conn.outlets:
                        outlets.append(outlet.label)
                    break

        return outlets

    def add_part(self, part):
        """
        Add a part to the graph.

        Parameters
        ----------
        part : BasePart
            Part to add.

        Raises
        ------
        ValueError
            If the part already exists in the graph.

        """
        if part.label in self._nodes:
            raise ValueError('part already exists')

        self._nodes[part.label] = part

        for conn in part.connections:
            for inlet in conn.inlets:
                edge = (inlet.label, part.label)
                edge_index = len(self._edges)
                if edge not in self._edges:
                    self._edges.append(edge)
                    self._edges_map[edge] = edge_index
                    self.states[edge_index] = StatePoint(self.states.fluid)

            for outlet in conn.outlets:
                edge = (part.label, outlet.label)
                edge_index = len(self._edges)
                if edge not in self._edges:
                    self._edges.append(edge)
                    self._edges_map[edge] = edge_index
                    self.states[edge_index] = StatePoint(self.states.fluid)

    def cloud_points(self, n=50, precise=False):
        """
        Generate cloud points between consecutive state points.

        Parameters
        ----------
        n : int, optional
            Number of points between each pair of state points.
            Default: 50
        precise : bool, optional
            Use a more precise version of interpolation by enthalpy
            instead of by temperature, slower. Default: False

        Returns
        -------
        list[StateCycle]
            List of state cycles containing the cloud points.

        """
        cloud = [StateCycle(self.states.fluid) for _ in range(self.points)]

        for i, edge1 in enumerate(self.edges):
            part_label = edge1[1]

            index1 = self.get_edge_index(edge1)
            state1 = self.states[index1]

            for outlet in self._nodes[part_label].get_outlets(edge1[0]):
                edge2 = (part_label, outlet.label)
                index2 = self.get_edge_index(edge2)
                state2 = self.states[index2]

                prop_x = next((prop.name for prop in self.states.constant_properties(index1, index2) if
                               prop.name not in ('S', 'Q', 'Y')), 'H' if precise else 'T')

                s_diff = state2['S'] - state1['S']
                x_diff = state2[prop_x] - state1[prop_x]

                s = np.linspace(state1['S'], state2['S'], n - 1)
                s = np.append(s, state2['S'])

                x = (x_diff / s_diff) * (s - state1['S']) + state1[prop_x] if abs(s_diff) > 1e-7 \
                    else np.linspace(state1[prop_x], state2[prop_x], n)

                for j in range(n):
                    cloud[i][j, 'S'] = s[j]
                    cloud[i][j, prop_x] = x[j]
                    cloud[i][j].properties(calc=['T'])
                    cloud[i][j].properties('T', 'S', exclude=['T'])
                    cloud[i][j, 'Y'] = state1['Y']

        return cloud



import numpy as np
from CoolProp.CoolProp import PropsSI
from itertools import combinations

from thercy.constants import Property, PropertyInfo


class StatePoint:
    """
    Thermodynamic state point.

    Properties
    ----------
    fluid : str
        Fluid composition or mixture name.
    data : list
        State data corresponding to thermodynamic properties.

    """

    def __init__(self, fluid, data=None):
        """
        Initialize a StatePoint instance.

        Parameters
        ----------
        fluid : str
            Fluid composition or mixture name.
        data : dict, optional
            State data corresponding to thermodynamic properties.
            Default: {}

        """
        self._fluid: str = fluid

        if data is None:
            data = {}
        self._data: list = [data.get(p.name) for p in Property]

    def __iter__(self):
        """Iterator over state data."""
        return iter(self._data)

    def __str__(self):
        """String representation of the state data."""
        return str(self._data)

    def __getitem__(self, prop):
        """
        Get the value of a property.

        Parameters
        ----------
        prop : Property | str | int
            Property to retrieve.

        Returns
        -------
        value : int
            State value of the property.

        """
        return self._data[PropertyInfo.get_intkey(prop)]

    def __setitem__(self, prop, value):
        """
        Set the value of a property.

        Parameters
        ----------
        prop : Property | str | int
            Property to set.
        value : float
            Value to assign to the property.

        """
        self._data[PropertyInfo.get_intkey(prop)] = value

    @property
    def fluid(self):
        """Get the fluid composition or mixture name."""
        return self._fluid

    @property
    def data(self):
        """Get the state data corresponding to thermodynamic properties."""
        return self._data

    def clone(self):
        """Create a copy of the state point."""
        data = {p.name: self._data[p.value] for p in Property}
        sp = StatePoint(self._fluid, data=data)
        return sp

    def from_array(self, x, exclude=None):
        """
        Set state data from an array.

        Parameters
        ----------
        x : iterable
            Array containing property values.
        exclude : list[str], optional
            Properties to exclude. Default: []

        Raises
        ------
        ValueError if `x` does not contain the correct amount of values.

        """
        if exclude is None:
            exclude = []

        if len(x) + len(exclude) != len(Property):
            raise ValueError('Not enough values to extract.')

        count = 0
        for prop in Property:
            if prop.name not in exclude:
                self._data[prop.value] = x[count]
                count += 1

    def to_array(self, exclude=None):
        """
        Convert state data to an array.

        Parameters
        ----------
        exclude : list[str], optional
            Properties to exclude. Default: []

        Returns
        -------
        x : numpy.ndarray
            Array containing property values.

        """
        if exclude is None:
            exclude = []

        length = len(Property) - len(exclude)
        x = np.zeros(length)
        count = 0

        for prop in Property:
            if prop.name not in exclude:
                x[count] = self._data[prop.value]
                count += 1

        return x

    def properties(self, prop1=None, prop2=None, calc=None, exclude=None):
        """
        Calculate missing properties based on known ones.

        Parameters
        ----------
        prop1 : Property | str | int, optional
            First known property. Default: None
        prop2 : Property | str | int, optional
            Second known property. Default: None
        calc : list[str], optional
            Properties to calculate. Default: all
        exclude : list[str], optional
            Properties to exclude from calculation. Default: None

        """
        # TODO: Move definition of not thermodynamic properties to `constants`
        not_thermo = ['Y']

        if calc is None:
            calc = [prop.name for prop in Property
                    if prop.name not in (prop1, prop2)]

        if exclude is None:
            exclude = []

        knowns = {Property(i).name: v for i, v in enumerate(self._data) if v is not None}

        if prop1 is not None and prop2 is not None:
            pairs = [(prop1, prop2)]
        elif prop1 is not None:
            pairs = [(prop1, p) for p in knowns.keys() if p != prop1 and p not in not_thermo]
        else:
            pairs = combinations(knowns.keys(), 2)

        to_calc = [prop for prop in calc
                   if prop not in not_thermo
                   and prop not in exclude]

        while len(to_calc) > 0:
            prop = to_calc.pop()

            for pair in pairs:
                if prop in pair:
                    continue

                try:
                    self[prop] = PropsSI(
                        prop,
                        pair[0],
                        knowns[pair[0]],
                        pair[1],
                        knowns[pair[1]],
                        self._fluid
                    )

                except ValueError as e:
                    pass



from abc import ABC, abstractmethod

from thercy.constants import PartType
from thercy.state import StateGraph, StatePoint


class Connection:
    def __init__(self, inlets, outlets):
        """
        Parameters
        ----------
        inlets : list[BasePart]
        outlets : list[BasePart]

        """
        self._inlets = inlets
        self._outlets = outlets

    @property
    def inlets(self):
        return self._inlets

    @property
    def outlets(self):
        return self._outlets


class BasePart(ABC):
    def __init__(self, label, type, connections=None):
        """
        Parameters
        ----------
        label : str
        type : PartType
        connections : list[Connection]

        """
        if connections is None:
            connections = []

        self._label: str = label
        self._type: PartType = type
        self._internal_conn: list[Connection] = connections

    @property
    def label(self):
        return self._label

    @property
    def type(self):
        return self._type

    @property
    def connections(self):
        return self._internal_conn

    @connections.setter
    def connections(self, connections):
        """
        Parameters
        ----------
        connections : list[Connection]

        """
        for con in connections:
            for inlet in con.inlets + con.outlets:
                if not isinstance(inlet, BasePart):
                    raise ValueError("part must be a BasePart")

        self._internal_conn = connections

    @property
    def inlet_parts(self):
        inlets = []
        for conn in self._internal_conn:
            for inlet in conn.inlets:
                inlets.append(inlet)

        return inlets

    @property
    def outlet_parts(self):
        outlets = []
        for conn in self._internal_conn:
            for outlet in conn.outlets:
                outlets.append(outlet)

        return outlets

    @property
    @abstractmethod
    def deltaH(self):
        raise NotImplementedError()

    def get_inlets(self, outlet: str):
        for conn in self._internal_conn:
            for outl in conn.outlets:
                if outl.label == outlet:
                    return conn.inlets

        return None

    def get_outlets(self, inlet: str):
        for conn in self._internal_conn:
            for inl in conn.inlets:
                if inl.label == inlet:
                    return conn.outlets

        return None

    def solve_conserv(self, inlets_state: dict[str, StatePoint], outlets_state: dict[str, StatePoint]):
        residual = [0.0, 0.0]

        for conn in self.connections:
            y_inl = 0.0
            h_inl = 0.0
            for label, state in inlets_state.items():
                if label in [p.label for p in conn.inlets]:
                    y_inl += state['Y']
                    h_inl += state['Y'] * state['H']

            y_outl = 0.0
            h_outl = 0.0
            for label, state in outlets_state.items():
                if label in [p.label for p in conn.outlets]:
                    y_outl += state['Y']
                    h_outl += state['Y'] * state['H']

            y = (y_inl + y_outl) / 2.0
            residual[0] += abs(y_outl - y_inl)
            residual[1] += h_outl - h_inl - y * self.deltaH

        return residual

    @abstractmethod
    def solve(self, inlets: dict[str, StatePoint]):
        raise NotImplementedError()



from thercy.state import StateGraph
from thercy.utils import list_like

from .rankine import Rankine
from .._parts import *


class RankineBuilder:
    _fluid: str
    _graph: StateGraph
    _parts: dict[str: BasePart]

    def __init__(self, fluid: str):
        self._connections: dict[str: list[(list[str], list[str])]] = {}
        self._fluid = fluid
        self._graph = StateGraph(fluid)
        self._parts: dict[str: BasePart] = {}

    def build(self):
        for label, part in self._parts.items():
            connections = []
            for conn in self._connections[label]:
                inlets = [self._parts[inlet] for inlet in conn[0]]
                outlets = [self._parts[outlet] for outlet in conn[1]]
                connections.append(Connection(inlets, outlets))

            part.connections = connections
            self._graph.add_part(part)

        return Rankine(self._fluid, self._graph)

    def add_condenser(self, label: str, inlet: str, outlet: str):
        condensator = Condenser(label)
        self._parts[label] = condensator
        self._connections[label] = [([inlet], [outlet])]

        return self

    def add_heater_closed(self, label: str, inlets_lp: list[str], inlet_hp: str, outlet_lp: str, outlet_hp: str, t_out: float):
        reheater_open = HeaterClosed(label, t_out)
        self._parts[label] = reheater_open
        self._connections[label] = [(inlets_lp, [outlet_lp]), ([inlet_hp], [outlet_hp])]

        return self

    def add_heater_open(self, label: str, inlets: list[str], outlet: str):
        reheater_open = HeaterOpen(label)
        self._parts[label] = reheater_open
        self._connections[label] = [(inlets, [outlet])]

        return self

    def add_pump(self, label: str, inlet: str, outlet: str, p_out: float, eta: float = 1.0):
        pump = Pump(label, p_out, eta=eta)
        self._parts[label] = pump
        self._connections[label] = [([inlet], [outlet])]

        return self

    def add_steam_generator(self, label: str, inlet: str, outlet: str, prop: str, value: float):
        heat_source = SteamGenerator(label, prop, value)
        self._parts[label] = heat_source
        self._connections[label] = [([inlet], [outlet])]

        return self

    def add_trap(self, label, inlet, outlet, p_out):
        """
        Parameters
        ----------
        label : str
        inlet : str
        outlet : str
        p_out : float

        """
        trap = Trap(label, p_out)
        self._parts[label] = trap
        self._connections[label] = [([inlet], [outlet])]

        return self

    def add_turbine(self, label, inlet, outlets, p_out, eta=1.0):
        """
        Parameters
        ----------
        label : str
        inlet : str
        outlets : str | list[str]
        p_out : float
        eta : float

        """
        turbine = Turbine(label, p_out, eta=eta)
        self._parts[label] = turbine

        _outlets = outlets if list_like(outlets) else [outlets]
        self._connections[label] = [([inlet], _outlets)]

        return self



from thercy.constants import PartType, PropertyInfo
from thercy.state import StatePoint

from .base_part import BasePart, Connection


class SteamGenerator(BasePart):
    _prop: str

    def __init__(self, label, prop, value, connections=None):
        """
        Parameters
        ----------
        label : str
        prop : str
        value : float
        connections : list[Connection]

        """
        super().__init__(
            label,
            PartType.HEAT_SOURCE,
            connections,
        )

        self._prop = PropertyInfo.get_strkey(prop)
        self._value = value
        self._deltaH = 0.0

    @property
    def deltaH(self):
        return self._deltaH

    def solve(self, inlets: dict[str, StatePoint]):
        outlets = {}

        inlet_label, inlet_state = next(iter(inlets.items()))
        outlet_state = inlet_state.clone()

        outlet_state[self._prop] = self._value
        outlet_state['P'] = inlet_state['P']
        outlet_state.properties(self._prop, 'P')

        self._deltaH = outlet_state['H'] - inlet_state['H']

        for outlet in self.get_outlets(inlet_label):
            outlets[outlet.label] = outlet_state.clone()

        return outlets



from enum import Enum


class Property(Enum):
    """
    Enumeration class defining thermodynamic properties.

    """
    T = 0  # Temperature
    P = 1  # Pressure
    D = 2  # Density
    H = 3  # Enthalpy
    S = 4  # Entropy
    Q = 5  # Quality
    Y = 6  # Fraction


class PropertyInfo:
    """
    Utility methods to retrieve information about thermodynamic properties.

    """
    _data = {
        'T': {'symbol': 'T', 'label': 'Temperature', 'unit': 'K'},
        'P': {'symbol': 'P', 'label': 'Pressure', 'unit': 'Pa'},
        'D': {'symbol': 'D', 'label': 'Density', 'unit': 'kg/m3'},
        'H': {'symbol': 'H', 'label': 'Enthalpy', 'unit': 'J/kg'},
        'S': {'symbol': 'S', 'label': 'Entropy', 'unit': 'J/kg/K'},
        'Q': {'symbol': 'Q', 'label': 'Quality', 'unit': '-'},
        'Y': {'symbol': 'Y', 'label': 'Fraction', 'unit': 'kg/m3'}
    }

    @classmethod
    def get_intkey(cls, prop):
        """
        Parameters
        ----------
        prop : Property | str | int
            Property.

        Returns
        -------
        key : int
            Integer value corresponding to the property `prop`.

        Raises
        ------
        TypeError if `prop` is not an valid property type.

        """
        if isinstance(prop, Property):
            key = prop.value
        elif isinstance(prop, str):
            key = getattr(Property, prop).value
        elif isinstance(prop, int):
            if prop > len(Property):
                raise ValueError('Property key is too large')
            key = prop
        else:
            raise TypeError('Unexpected key type')

        return key

    @classmethod
    def get_strkey(cls, prop):
        """
        Parameters
        ----------
        prop : Property | str | int
            Property.

        Returns
        -------
        key : str
            String corresponding to the property `prop`.

        Raises
        ------
        TypeError if `prop` is not an valid property type.

        """
        if isinstance(prop, Property):
            key = prop.name
        elif isinstance(prop, str):
            if prop not in cls._data.keys():
                raise ValueError('Invalid property name')
            key = prop
        elif isinstance(prop, int):
            key = Property(prop).name
        else:
            raise TypeError('Unexpected key type')

        return key

    @classmethod
    def symbol(cls, prop):
        """
        Parameters
        ----------
        prop : Property | str | int
            Property.

        Returns
        -------
        symbol : str
            Symbol corresponding to the property `prop`.

        """
        return cls._data[cls.get_strkey(prop)]['symbol']

    @classmethod
    def label(cls, prop):
        """
        Parameters
        ----------
        prop : Property | str | int
            Property.

        Returns
        -------
        label : str
            Label or description corresponding to the property `prop`.

        """
        return cls._data[cls.get_strkey(prop)]['label']

    @classmethod
    def unit(cls, prop):
        """
        Parameters
        ----------
        prop : Property | str | int
            Property.

        Returns
        -------
        unit : str
            SI unit corresponding to the property `prop`.

        """
        return cls._data[cls.get_strkey(prop)]['unit']


class PartType(Enum):
    """
    Enumeration class defining types of parts in a thermodynamic cycle.

    """
    CONDENSATOR = 0
    HEAT_SOURCE = 1
    PUMP = 2
    REHEATER_CLOSE = 3
    REHEATER_OPEN = 4
    TURBINE = 5



import numpy as np
from scipy.optimize import minimize, root

from thercy.constants import PartType, Property
from thercy.state import StateGraph
from thercy.utils import norm_l1, norm_l2, norm_lmax, norm_lp


class Rankine:
    def __init__(self, fluid: str, parts: StateGraph):
        self._graph: StateGraph = parts
        self._fluid: str = fluid
        self._heat_input: float = 0.0
        self._heat_output: float = 0.0
        self._work_pumps: float = 0.0
        self._work_turbines: float = 0.0

    def __len__(self):
        return len(self._graph)

    def __str__(self):
        return str(self._graph)

    @property
    def graph(self):
        return self._graph

    @property
    def states(self):
        return self._graph.states

    def _equation_thermo(self, x: np.ndarray):
        len_props = len(Property)
        len_states = self._graph.points
        residual = np.zeros_like(x)

        for index in range(len_states):
            index_begin = index * len_props
            index_end = index_begin + len_props
            self._graph.states[index].from_array(x[index_begin:index_end])

        for part in self._graph.nodes.values():
            inlets_state = {p.label: self._graph.get_state((p.label, part.label)) for p in part.inlet_parts}
            sol = self._graph[part.label].solve(inlets_state)

            for label_outlet, value in sol.items():
                edge = (part.label, label_outlet)
                edge_index = self._graph.get_edge_index(edge)
                for prop in Property:
                    if value[prop.value] is not None:
                        self._graph.states[edge_index][prop.value] = value[prop.value]

        for index in range(len_states):
            index_begin = index * len_props
            index_end = index_begin + len_props
            residual[index_begin:index_end - 1] = (x[index_begin:index_end - 1]
                                                   - self._graph.states[index].to_array(['Y']))

        return residual

    def _equation_conserv(self, y: np.ndarray):
        residual = np.zeros(2 * len(self._graph))

        for index in range(self._graph.points):
            self._graph.states[index]['Y'] = y[index]

        for i, part in enumerate(self._graph.nodes.values()):
            inlets_state = {p.label: self._graph.get_state((p.label, part.label)) for p in part.inlet_parts}
            outlets_state = {p.label: self._graph.get_state((part.label, p.label)) for p in part.outlet_parts}
            residual[2 * i:2 * i + 2] = self._graph[part.label].solve_conserv(inlets_state, outlets_state)

        return residual

    def _iterate_conserv_scalar(self, x: np.ndarray, verbose=0):
        residual = self._equation_conserv(x)
        residual_mass = norm_l2(residual[0::2])
        residual_energy = norm_l2(residual[1::2])

        if verbose >= 3:
            print(f"{'Rankine._iterate_conserv : ':40s}"
                  f"{residual_energy:3e} | {np.sqrt(residual_energy / len(self._graph)):3e} | "
                  f"{residual_mass:.3e} | {np.sqrt(residual_mass / len(self._graph)):.3e}")

        return residual_energy + 1e9 * residual_mass

    def _iterate_thermo(self, x0: np.ndarray, xtol=1e-4, maxfev=10, verbose=0):
        sol = root(
            self._equation_thermo,
            x0,
            method='df-sane',
            options={'fatol': xtol, 'maxfev': maxfev}
        )

        len_props = len(Property)
        len_states = self._graph.points
        for index in range(len_states):
            index_begin = index * len_props
            index_end = index_begin + len_props
            self._graph.states[index].from_array(sol.x[index_begin:index_end])

        if verbose >= 3:
            print(f"{'Rankine._iterate_thermo : ':40s}"
                  f"fun = {sol.fun}, nfev={sol.nfev}")

        return sol.x

    def _iterate(self, x0_fraction: np.ndarray, x0_states: np.ndarray, update_states=False, xtol=1e-4, verbose=0):
        x0_states_ = x0_states.copy()
        x0_states_[len(Property) - 1::len(Property)] = x0_fraction
        sol_thermo = self._iterate_thermo(x0_states_, xtol=xtol, verbose=verbose)

        bounds = self._graph.points * [(1.0, 1000.0)]
        sol_conserv = minimize(
            self._iterate_conserv_scalar,
            bounds=bounds,
            x0=x0_fraction,
            args=(verbose,),
            method='L-BFGS-B',
            options={'ftol': xtol / 10, 'maxiter': 1000}
        )

        if update_states:
            x0_states = x0_states_

        if verbose >= 2:
            # residual = self._equation_conserv(sol_conserv.x)
            # residual_mass = norm_l2(residual[0::2])
            # residual_energy = norm_l2(residual[1::2])
            # print(f"{'Rankine._iterate : ':40s}"
            #       f"{residual_energy:3e} | {np.sqrt(residual_energy / 11):3e} | "
            #       f"{residual_mass:.3e} | {np.sqrt(residual_mass / 11):.3e}")
            pass

        return x0_fraction - sol_conserv.x

    def solve(self, x0, knowns=None, xtol=1e-4, verbose=0):
        if knowns is None:
            raise ValueError('At least one property must be known for every state')

        len_props = len(Property)
        len_props_input = len(knowns)
        len_exclude = len_props - len_props_input
        x0_complete = np.ones(len_props * len(x0) // len_props_input)

        for i in range(0, len(x0), len_props_input):
            new_index = i + (i // len_props_input) * len_exclude
            x0_complete[new_index:new_index + len_props_input] = x0[i:i + len_props_input]

        if 'Y' in knowns:
            x0_fraction = x0_complete[len_props - 1::len_props]
            x0_fraction *= 1000.0 / np.max(x0_fraction)
        else:
            x0_fraction = np.full(self._graph.points, 1000.0)
            x0_complete[len_props - 1::len_props] = x0_fraction

        # Solve for the conservation laws
        sol_conserv = root(
            self._iterate,
            x0_fraction,
            method='df-sane',
            args=(x0_complete, True, xtol, verbose),
            options={'sigma_0': 1e-4, 'fatol': xtol, 'ftol': 0.0, 'maxfev': 100}
        )

        # Process the solution, guarantees that graph.states contains the
        # correct values
        sol_x = x0_complete.copy()
        sol_x[len_props - 1::len_props] = sol_conserv.x * 1000.0 / np.max(sol_conserv.x)
        sol_thermo = root(
            self._equation_thermo,
            sol_x,
            method='df-sane',
            options={'fatol': xtol, 'maxfev': 10}
        )
        sol_x = sol_thermo.x
        for index in range(self._graph.points):
            index_begin = index * len_props
            index_end = index_begin + len_props
            self._graph.states[index].from_array(sol_x[index_begin:index_end])

        # Post-processing
        for part in self._graph.nodes.values():
            y_part = sum(self._graph.states[self._graph.get_edge_index((inlet.label, part.label))]['Y']
                         for inlet in part.inlet_parts)

            match part.type:
                case PartType.CONDENSATOR:
                    self._heat_output += y_part * part.deltaH
                case PartType.HEAT_SOURCE:
                    self._heat_input += y_part * part.deltaH
                case PartType.PUMP:
                    self._work_pumps += y_part * part.deltaH
                case PartType.TURBINE:
                    self._work_turbines += y_part * part.deltaH

        return self._graph.states

    @property
    def bwr(self):
        return self._work_pumps / -self._work_turbines

    @property
    def cycle(self):
        return self._cycle

    @property
    def efficiency(self):
        return self.work / self._heat_input

    @property
    def heat_input(self):
        return self._heat_input

    @property
    def heat_output(self):
        return -self._heat_output

    def massflow(self, power):
        return power / self.work

    @property
    def work(self):
        return -(self._work_pumps + self._work_turbines)



import numpy as np


def list_like(value):
    """
    Check if a variable is list-like.

    Parameters
    ----------
    value : any
        Variable to check.

    Returns
    -------
    bool
        `True` if the variable is list-like, `False` otherwise.

    """
    return (not hasattr(value, "strip")
            and (hasattr(value, "__getitem__")
                 or hasattr(value, "__iter__")))


def norm_l1(x: np.ndarray, rescale=False):
    norm = np.sum(np.abs(x))

    if rescale:
        norm /= len(x)

    return norm


def norm_l2(x: np.ndarray, rescale=False):
    norm = np.sum(np.square(x))

    if rescale:
        norm = np.sqrt(norm / len(x))

    return norm


def norm_lmax(x: np.ndarray, rescale=False):
    norm = np.max(np.abs(x))

    if not rescale:
        norm *= len(x)

    return norm


def norm_lp(x: np.ndarray, p: float, rescale=False):
    norm = np.sum(np.abs(x ** p))

    if rescale:
        norm = (norm / len(x)) ** (1.0 / p)

    return norm



from .base_part import BasePart, Connection
from .condenser import Condenser
from .heater_closed import HeaterClosed
from .heater_open import HeaterOpen
from .pump import Pump
from .steam_generator import SteamGenerator
from .trap import Trap
from .turbine import Turbine
